/*
 *  Copyright (c) 2007-2008 Jiri Benc <jbenc@upir.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <gtk/gtk.h>
#include <glib-object.h>
#include <libhildondesktop/libhildondesktop.h>
#include <libhildondesktop/hildon-desktop-toggle-button.h>
#include "cmdnavigator.h"
#include "cmdnav.h"
#include "prefs.h"
#include "run.h"

#define MENU_HEIGHT	13
#define DEBUG

#ifdef DEBUG
#include <stdio.h>
#define debug_print	printf
#else
#define debug_print(...)
#endif

HD_DEFINE_PLUGIN(CmdNavigatorPlugin, cmd_navigator_plugin, TASKNAVIGATOR_TYPE_ITEM);

typedef struct _CmdNavigatorPluginPrivate CmdNavigatorPluginPrivate;
struct _CmdNavigatorPluginPrivate {
	GtkWidget *button;
	GtkWidget *menu;
	struct cmd *cmds;
	gboolean valid;
	guint prefs_handler;
};

#define CMD_NAVIGATOR_PLUGIN_GET_PRIVATE(obj)	(G_TYPE_INSTANCE_GET_PRIVATE((obj), TYPE_CMD_NAVIGATOR_PLUGIN, CmdNavigatorPluginPrivate))

/* forward declarations */
static void cmd_navigator_plugin_finalize(GObject *obj);
static void menu_destroy(CmdNavigatorPluginPrivate *priv);
static void menu_invalidate(void *data);
static gboolean menu_on(GtkToggleButton *button, gpointer data);


static void cmd_navigator_plugin_class_init(CmdNavigatorPluginClass *klass)
{
	GObjectClass *object_class = (GObjectClass *)klass;

	object_class->finalize = cmd_navigator_plugin_finalize;

	g_type_class_add_private(object_class, sizeof(CmdNavigatorPluginPrivate));
}

static void cmd_navigator_plugin_init(CmdNavigatorPlugin *plugin)
{
	CmdNavigatorPluginPrivate *priv;

	debug_print("cmdnav: init\n");

	if (g_file_test("/media/mmc1/disable-cmd-navigator", G_FILE_TEST_EXISTS) ||
			g_file_test("/media/mmc2/disable-cmd-navigator", G_FILE_TEST_EXISTS))
		return;

	priv = CMD_NAVIGATOR_PLUGIN_GET_PRIVATE(plugin);
	plugin->private_data = priv;

	priv->prefs_handler = init_prefs(menu_invalidate, priv);

	priv->button = hildon_desktop_toggle_button_new();
	gtk_container_add(GTK_CONTAINER(priv->button),
		gtk_image_new_from_file("/usr/share/icons/hicolor/64x64/apps/terminal.png"));
//	gtk_widget_set_size_request(priv->button, 80, 80);
	gtk_widget_show_all(priv->button);
	gtk_container_add(GTK_CONTAINER(plugin), priv->button);

	g_signal_connect(G_OBJECT(priv->button), "toggled", G_CALLBACK(menu_on), priv);
}

static void cmd_navigator_plugin_finalize(GObject *obj)
{
	CmdNavigatorPlugin *plugin = CMD_NAVIGATOR_PLUGIN(obj);
	CmdNavigatorPluginPrivate *priv = (CmdNavigatorPluginPrivate *)plugin->private_data;

	debug_print("cmdnav: finalize\n");

	if (priv) {
		menu_destroy(priv);
		/* button is destroyed with the container */
		deinit_prefs(priv->prefs_handler);
	}

	G_OBJECT_CLASS(cmd_navigator_plugin_parent_class)->finalize(obj);
}


static void menu_invalidate(void *data)
{
	CmdNavigatorPluginPrivate *priv = data;

	debug_print("cmdnav: invalidate\n");

	priv->valid = FALSE;
}

static void menu_destroy(CmdNavigatorPluginPrivate *priv)
{
	if (priv->menu) {
		gtk_menu_shell_deactivate(GTK_MENU_SHELL(priv->menu));
		gtk_widget_destroy(priv->menu);
		priv->menu = NULL;
	}
	if (priv->cmds) {
		free_prefs(priv->cmds);
		priv->cmds = NULL;
	}
}

static void menu_position(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, gpointer data)
{
	CmdNavigatorPluginPrivate *priv = data;
	GtkRequisition req;
	int screen_dim;

	gtk_widget_size_request(GTK_WIDGET(menu), &req);

	gdk_window_get_origin(priv->button->window, x, y);
	*x += priv->button->allocation.width - 1;
	if (GTK_WIDGET_NO_WINDOW(priv->button)) {
		*x += priv->button->allocation.x;
		*y += priv->button->allocation.y;

		/* move the menu up if it's too high */
		screen_dim = gdk_screen_get_height(gtk_widget_get_screen(GTK_WIDGET(menu)));
		if (req.height + *y > screen_dim) {
			*y = screen_dim - req.height;
			if (*y < 0)
				*y = 0;
		}
	}

	/* reduce the width of the menu if it's too large */
	screen_dim = gdk_screen_get_width(gtk_widget_get_screen(GTK_WIDGET(menu)));
	if (req.width + *x > screen_dim)
		gtk_widget_set_size_request(GTK_WIDGET(menu), screen_dim - *x, req.height);

	*push_in = FALSE;
}

static void menu_off(GtkMenu *menu, gpointer data)
{
	CmdNavigatorPluginPrivate *priv = data;

	(void)menu;
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(priv->button), FALSE);
}

static void run_one_cmd(GtkMenuItem *item, gpointer data)
{
	(void)item;
	run_cmd(data);
}

static void menu_create(CmdNavigatorPluginPrivate *priv)
{
	struct cmd *cur;
	GtkWidget *item, *label;
	int cnt, height, x, y;

	if (priv->valid)
		return;

	menu_destroy(priv);
	priv->menu = gtk_menu_new();
	priv->cmds = get_prefs(&cnt);

	priv->valid = TRUE;

	x = (cnt + 1 + MENU_HEIGHT - 1) / MENU_HEIGHT;
	height = (cnt + 1 + x - 1) / x;
	x = y = 0;
	for (cur = priv->cmds->next; cur; cur = cur->next) {
		item = gtk_menu_item_new_with_label(cur->name);
		g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(run_one_cmd), cur);
		gtk_menu_attach(GTK_MENU(priv->menu), item, x, x + 1, y, y + 1);
		if (++y == height) {
			y = 0;
			x++;
		}
	}

	item = gtk_menu_item_new();
	label = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(label), "<span foreground=\"blue\" style=\"italic\">Edit menu...</span>");
	gtk_container_add(GTK_CONTAINER(item), label);
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(run_prefs), NULL);
	gtk_menu_attach(GTK_MENU(priv->menu), item, x, x + 1, y, y + 1);

	gtk_widget_show_all(priv->menu);

	g_signal_connect(G_OBJECT(priv->menu), "hide", G_CALLBACK(menu_off), priv);
}

static gboolean menu_on(GtkToggleButton *button, gpointer data)
{
	CmdNavigatorPluginPrivate *priv = data;

	debug_print("cmdnav: toggle\n");

	dezombie();
	if (!gtk_toggle_button_get_active(button))
		return TRUE;

	menu_create(priv);
	gtk_menu_popup(GTK_MENU(priv->menu), NULL, NULL, menu_position, data,
		1, gtk_get_current_event_time());
	return TRUE;
}
