/*
 *  Copyright (c) 2007-2008 Jiri Benc <jbenc@upir.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <libosso.h>
#include "cmdnavigator.h"
#include "prefs.h"

static GtkTreeView *view;
static GtkWindow *prefwin;

#ifdef DEBUG
static void debug_print_prefs(struct cmd *cmds)
{
	struct cmd *cur;

	g_print("Prefs dump:\n");
	for (cur = cmds->next; cur; cur = cur->next) {
		g_print("\t");
		if (cur->flags & CMD_FLAG_DELETED)
			g_print("[DEL] ");
		if (cur->flags & CMD_FLAG_CHANGED)
			g_print("[CHN] ");
		if (!cur->id)
			g_print("[NEW] ");
		else
			g_print("(%d) ", cur->id);
		g_print("%s => %s\n", cur->name, cur->cmd);
	}
	g_print("End prefs dump.\n");
}
#else
static inline void debug_print_prefs(struct cmd *cmds) { (void)cmds; }
#endif

static struct cmd *cmd_add(struct cmd *cmds, int index)
{
	struct cmd *cur, *last = NULL;
	int i;

	i = 0;
	cur = cmds;
	while (cur && i <= index) {
		last = cur;
		cur = cur->next;
		if (cur && !(cur->flags & CMD_FLAG_DELETED))
			i++;
	}
	cur = g_malloc0(sizeof(struct cmd));
	/* in OS2008: cur = g_slice_alloc0(sizeof(struct cmd)); */
	cur->next = last->next;
	last->next = cur;
	return cur;
}

static void cmd_del(struct cmd *cmds, int index)
{
	struct cmd *cur, *last = NULL;
	int i;

	i = 0;
	cur = cmds;
	while (cur && i <= index) {
		last = cur;
		cur = cur->next;
		if (cur && !(cur->flags & CMD_FLAG_DELETED))
			i++;
	}
	if (cur) {
		cur->flags |= CMD_FLAG_DELETED;
		/* deleted items are always at the beginning */
		last->next = cur->next;
		cur->next = cmds->next;
		cmds->next = cur;
	}
}

static struct cmd *cmd_get(struct cmd *cmds, int index)
{
	struct cmd *cur;
	int i;

	i = 0;
	cur = cmds;
	while (cur && i <= index) {
		cur = cur->next;
		if (cur && !(cur->flags & CMD_FLAG_DELETED))
			i++;
	}
	return cur;
}

static void cmd_set_changed(struct cmd *cmd)
{
	cmd->flags |= CMD_FLAG_CHANGED;
}

/* cmd_move_down is equal to cmd_move_up(cmds, index + 1) */
static gboolean cmd_move_up(struct cmd *cmds, int index)
{
	struct cmd *cur, *last = NULL;
	int i;

	if (index <= 0)
		return FALSE;
	i = 0;
	cur = cmds;
	while (cur && i < index) {
		last = cur;
		cur = cur->next;
		if (cur && !(cur->flags & CMD_FLAG_DELETED))
			i++;
	}
	if (cur && cur->next) {
		last->next = cur->next;
		last = cur->next->next;
		cur->next->next = cur;
		cur->next = last;
		return TRUE;
	}
	return FALSE;
}

static GtkTreeView *prefs_make_view(struct cmd *cmds)
{
	GtkListStore *store;
	GtkTreeIter iter;
	GtkTreeView *res;

	store = gtk_list_store_new(1, G_TYPE_STRING);
	for (cmds = cmds->next; cmds; cmds = cmds->next) {
		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter, 0, cmds->name, -1);
	}
	res = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(store)));
	gtk_tree_view_append_column(res,
		gtk_tree_view_column_new_with_attributes(NULL, gtk_cell_renderer_text_new(),
			"text", 0, NULL));
	return res;
}

static int get_view_pos(void)
{
	GtkTreePath *tmp;
	gint *pos;

	gtk_tree_view_get_cursor(view, &tmp, NULL);
	if (!tmp)
		return -1;
	pos = gtk_tree_path_get_indices(tmp);
	return pos ? pos[0] : -1;
}

static void set_view_pos(int pos)
{
	GtkTreePath *path;

	path = gtk_tree_path_new_from_indices(pos, -1);
	gtk_tree_view_set_cursor(view, path, NULL, FALSE);
	gtk_tree_path_free(path);
}

static void get_iter_from_pos(GtkListStore *store, GtkTreeIter *iter, int pos)
{
	GtkTreePath *path;

	path = gtk_tree_path_new_from_indices(pos, -1);
	gtk_tree_model_get_iter(GTK_TREE_MODEL(store), iter, path);
	gtk_tree_path_free(path);
}

static void dialog_add_edit_cmd(struct cmd *cmds, gboolean edit)
{
	GtkDialog *dialog;
	GtkWidget *button, *table;
	GtkWidget *entry_name, *entry_cmd, *check_xterm, *check_root;
	int pos, res;
	struct cmd *cmd_item = NULL;

	pos = get_view_pos();
	if (edit) {
		if (pos >= 0)
			cmd_item = cmd_get(cmds, pos);
		if (pos < 0 || !cmd_item) {
			gtk_widget_grab_focus(GTK_WIDGET(view));
			return;
		}
	}

	dialog = GTK_DIALOG(gtk_dialog_new());
	if (!dialog)
		return;
	gtk_window_set_transient_for(GTK_WINDOW(dialog), prefwin);
	gtk_window_set_title(GTK_WINDOW(dialog), edit ? "Edit command" : "Add command");
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);

	button = gtk_button_new_with_label("OK");
	gtk_dialog_add_action_widget(dialog, button, GTK_RESPONSE_ACCEPT);
	GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_dialog_add_action_widget(dialog, gtk_button_new_with_label("Cancel"), GTK_RESPONSE_REJECT);

	table = gtk_table_new(2, 4, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 5);
	gtk_table_attach_defaults(GTK_TABLE(table), gtk_label_new("Description:"), 0, 1, 0, 1);
	entry_name = gtk_entry_new();
	gtk_entry_set_activates_default(GTK_ENTRY(entry_name), TRUE);
	gtk_widget_set_size_request(entry_name, 400, -1);
	gtk_table_attach_defaults(GTK_TABLE(table), entry_name, 1, 2, 0, 1);
	gtk_table_attach_defaults(GTK_TABLE(table), gtk_label_new("Command:"), 0, 1, 1, 2);
	entry_cmd = gtk_entry_new();
	gtk_entry_set_activates_default(GTK_ENTRY(entry_cmd), TRUE);
	hildon_gtk_entry_set_input_mode(GTK_ENTRY(entry_cmd), HILDON_GTK_INPUT_MODE_FULL);	/* do not capitalize */
	gtk_table_attach_defaults(GTK_TABLE(table), entry_cmd, 1, 2, 1, 2);
	check_xterm = gtk_check_button_new_with_label("run in XTerm");
	gtk_table_attach_defaults(GTK_TABLE(table), check_xterm, 1, 2, 2, 3);
	check_root = gtk_check_button_new_with_label("run as root");
	gtk_table_attach_defaults(GTK_TABLE(table), check_root, 1, 2, 3, 4);

	if (edit) {
		gtk_entry_set_text(GTK_ENTRY(entry_name), cmd_item->name);
		gtk_entry_set_text(GTK_ENTRY(entry_cmd), cmd_item->cmd);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_xterm), cmd_item->flags & CMD_FLAG_XTERM);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check_root), cmd_item->flags & CMD_FLAG_ROOT);
	}

	gtk_container_add(GTK_CONTAINER(dialog->vbox), table);

	gtk_widget_show_all(GTK_WIDGET(dialog));
	res = gtk_dialog_run(dialog);
	if (res == GTK_RESPONSE_ACCEPT) {
		GtkListStore *store;
		GtkTreeIter iter;

		if (edit) {
			cmd_set_changed(cmd_item);
			g_free(cmd_item->name);
			g_free(cmd_item->cmd);
		} else
			cmd_item = cmd_add(cmds, pos + 1);
		cmd_item->name = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_name)));
		cmd_item->cmd = g_strdup(gtk_entry_get_text(GTK_ENTRY(entry_cmd)));
		cmd_item->flags = (cmd_item->flags & ~(CMD_FLAG_XTERM | CMD_FLAG_ROOT)) |
			(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check_xterm)) ? CMD_FLAG_XTERM : 0) |
			(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(check_root)) ? CMD_FLAG_ROOT : 0);

		store = GTK_LIST_STORE(gtk_tree_view_get_model(view));
		if (edit)
			get_iter_from_pos(store, &iter, pos);
		else
			gtk_list_store_insert(store, &iter, pos + 1);
		gtk_list_store_set(store, &iter, 0, cmd_item->name, -1);
		if (!edit)
			set_view_pos(pos + 1);
	}

	gtk_widget_destroy(GTK_WIDGET(dialog));
	gtk_widget_grab_focus(GTK_WIDGET(view));
	debug_print_prefs(cmds);
}

static void dialog_add_cmd(GtkButton *clicked, struct cmd *cmds)
{
	(void)clicked;
	dialog_add_edit_cmd(cmds, FALSE);
}

static void dialog_edit_cmd(GtkButton *clicked, struct cmd *cmds)
{
	(void)clicked;
	dialog_add_edit_cmd(cmds, TRUE);
}

static void dialog_delete_cmd(GtkButton *clicked, struct cmd *cmds)
{
	int pos;
	GtkListStore *store;
	GtkTreeIter iter;

	(void)clicked;
	pos = get_view_pos();
	if (pos < 0) {
		gtk_widget_grab_focus(GTK_WIDGET(view));
		return;
	}
	cmd_del(cmds, pos);

	store = GTK_LIST_STORE(gtk_tree_view_get_model(view));
	get_iter_from_pos(store, &iter, pos);
	gtk_list_store_remove(store, &iter);
	set_view_pos(pos);

	gtk_widget_grab_focus(GTK_WIDGET(view));
	debug_print_prefs(cmds);
}

static void dialog_move_up_down(struct cmd *cmds, gboolean down)
{
	int pos;
	GtkListStore *store;
	GtkTreeIter iter1, iter2;

	pos = get_view_pos();
	if (down && pos >= 0)
		pos++;
	if (pos < 0 || !cmd_move_up(cmds, pos)) {
		gtk_widget_grab_focus(GTK_WIDGET(view));
		return;
	}

	store = GTK_LIST_STORE(gtk_tree_view_get_model(view));
	get_iter_from_pos(store, &iter1, pos);
	get_iter_from_pos(store, &iter2, pos - 1);
	gtk_list_store_move_before(store, &iter1, &iter2);
	if (down)
		set_view_pos(pos);
	else
		set_view_pos(pos - 1);

	gtk_widget_grab_focus(GTK_WIDGET(view));
	debug_print_prefs(cmds);
}

static void dialog_move_up(GtkButton *clicked, struct cmd *cmds)
{
	(void)clicked;
	dialog_move_up_down(cmds, FALSE);
}

static void dialog_move_down(GtkButton *clicked, struct cmd *cmds)
{
	(void)clicked;
	dialog_move_up_down(cmds, TRUE);
}

static void run_prefs(GtkWindow *parent)
{
	struct cmd *cmds;
	GtkDialog *dialog;
	GtkWidget *button, *scroll, *hbox, *vbox;
	int res;

	preload_prefs();
	cmds = _get_prefs(NULL);

	dialog = GTK_DIALOG(gtk_dialog_new());
	if (!dialog)
		return;
	gtk_window_set_transient_for(GTK_WINDOW(dialog), parent);
	gtk_window_set_title(GTK_WINDOW(dialog), "Edit commands");
	gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);

	prefwin = GTK_WINDOW(dialog);
	view = prefs_make_view(cmds);

	gtk_dialog_add_action_widget(dialog, gtk_button_new_with_label("OK"), GTK_RESPONSE_ACCEPT);
	gtk_dialog_add_action_widget(dialog, gtk_button_new_with_label("Cancel"), GTK_RESPONSE_REJECT);

	scroll = gtk_scrolled_window_new(NULL, GTK_ADJUSTMENT(gtk_adjustment_new(0, 0, 25, 0, 0, 0)));
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
	gtk_widget_set_size_request(GTK_WIDGET(scroll), 400, -1);
	gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(view));

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), scroll, TRUE, TRUE, 0);
	vbox = gtk_vbutton_box_new();
	gtk_box_set_spacing(GTK_BOX(vbox), 3);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

	button = gtk_button_new_with_label("Edit");
	gtk_container_add(GTK_CONTAINER(vbox), button);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(dialog_edit_cmd), cmds);
	button = gtk_button_new_with_label("Add");
	gtk_container_add(GTK_CONTAINER(vbox), button);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(dialog_add_cmd), cmds);
	button = gtk_button_new_with_label("Delete");
	gtk_container_add(GTK_CONTAINER(vbox), button);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(dialog_delete_cmd), cmds);
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_icon_name("qgn_indi_arrow_up", GTK_ICON_SIZE_BUTTON));
	gtk_container_add(GTK_CONTAINER(vbox), button);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(dialog_move_up), cmds);
	button = gtk_button_new();
	gtk_button_set_image(GTK_BUTTON(button), gtk_image_new_from_icon_name("qgn_indi_arrow_down", GTK_ICON_SIZE_BUTTON));
	gtk_container_add(GTK_CONTAINER(vbox), button);
	g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(dialog_move_down), cmds);

	gtk_container_add(GTK_CONTAINER(dialog->vbox), hbox);

	gtk_widget_show_all(GTK_WIDGET(dialog));
	res = gtk_dialog_run(dialog);
	if (res == GTK_RESPONSE_ACCEPT) {
		set_prefs(cmds);
	}

	gtk_widget_destroy(GTK_WIDGET(dialog));
	free_prefs(cmds);
}

osso_return_t execute(osso_context_t * osso, gpointer data)
{
	guint handle;

	(void)osso;
	handle = init_prefs(NULL, NULL);
	run_prefs(GTK_WINDOW(data));
	deinit_prefs(handle);
	return 0;
}
